/***************************************************************************
 $RCSfile$
 -------------------
 cvs         : $Id: kbd_handler.c 17 2005-11-29 18:26:56Z aquamaniac $
 begin       : Mon Mar 01 2004
 copyright   : (C) 2004 by Martin Preuss
 email       : martin@libchipcard.de

 ***************************************************************************
 *          Please see toplevel file COPYING for license details           *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "kbd_handler_p.h"

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <assert.h>

//#define DEBUG_MSGS



KBD_HANDLER *kbd_handler_new() {
  KBD_HANDLER *kh;

  kh=(KBD_HANDLER*)malloc(sizeof(KBD_HANDLER));
  assert(kh);
  memset(kh, 0, sizeof(KBD_HANDLER));

  return kh;
}



void kbd_handler_free(KBD_HANDLER *kh) {
  if (kh) {
    if (kh->private_data && kh->freeDataFn)
      kh->freeDataFn(kh, kh->private_data);
    free((void*)kh);
  }
}



int kbd_handler_init(KBD_HANDLER *kh) {
  if (kh->initFn)
    return kh->initFn(kh);
  else
    return 0;
}



void kbd_handler_close(KBD_HANDLER *kh) {
  if (kh->closeFn)
    kh->closeFn(kh);
}



int kbd_handler_handle(KBD_HANDLER *kh) {
  if (kh->handleFn)
    return kh->handleFn(kh);
  else
    return 0;
}



int kbd_handler_get_card(KBD_HANDLER *kh,
                         unsigned char *buffer,
                         unsigned int *bufsize) {
  assert(kh);
  assert(kh->getCardFn);
  return kh->getCardFn(kh, buffer, bufsize);
}



void kbd_handler_set_private_data(KBD_HANDLER *kh, void *p,
                                  KBD_HANDLER_FREE_PRIVATE_DATA_FN fn) {
  assert(kh);
  if (kh->private_data && kh->freeDataFn)
    kh->freeDataFn(kh, p);
  kh->private_data=p;
  kh->freeDataFn=fn;
}



void *kbd_handler_get_private_data(KBD_HANDLER *kh) {
  assert(kh);
  return kh->private_data;
}



void kbd_handler_set_init_fn(KBD_HANDLER *kh,
                             KBD_HANDLER_INIT_FN fn) {
  assert(kh);
  kh->initFn=fn;
}



void kbd_handler_set_close_fn(KBD_HANDLER *kh,
                              KBD_HANDLER_CLOSE_FN fn) {
  assert(kh);
  kh->closeFn=fn;
}



void kbd_handler_set_handle_fn(KBD_HANDLER *kh,
                               KBD_HANDLER_HANDLE_FN fn) {
  assert(kh);
  kh->handleFn=fn;
}



void kbd_handler_set_get_card_fn(KBD_HANDLER *kh,
                                 KBD_HANDLER_GET_CARD_FN fn) {
  assert(kh);
  kh->getCardFn=fn;
}



int kbd_handler_make_card_data(KBD_HANDLER *kh,
                               const unsigned char *ibuf,
                               unsigned int ilen,
                               unsigned char *buffer,
                               unsigned int *buflen) {
  static char header[30]= {
    0xa2, 0x13, 0x10, 0x91, 0x46, 0x0b, 0x82, 0x05,
    0x44, 0x45, 0x50, 0x41, 0x56, 0xff, 0xff, 0xff,
    0xff, 0x61, 0x0b, 0x4f, 0x06, 0xd2, 0x76, 0x00,
    0x00, 0x01, 0x01, 0x53, 0x01, 0x00 };
  static unsigned char lineTags[KBD_HANDLER_CARD_LINES]= {
    0x80, 0x81, 0x8f, 0x82, 0x83, 0x90, 0x84, 0x85,
    0x86, 0x87, 0x88, 0x89, 0x8a, 0x8b, 0x8c, 0x8d };
  unsigned char *d;
  unsigned char *dstart;
  const unsigned char *p;
  int i;
  unsigned int bytesLeft;
  unsigned int j;

  if (*buflen<256) {
    fprintf(stderr, "ERROR: Buffer too small.\n");
    return KBD_HANDLER_ERROR_BUFFER_TOO_SMALL;
  }

  memmove(buffer, header, sizeof(header));
  d=dstart=buffer+sizeof(header);

  p=ibuf;

  *(d++)=0x60;
  *(d++)=0x00; /* we set this later */
  j=2;
  bytesLeft=*buflen-j-sizeof(header);
  for (i=0; i<KBD_HANDLER_CARD_LINES; i++) {
    if (!*p)
      break;
    else {
      unsigned int k;
      unsigned char l;
      const unsigned char *pbegin;

      pbegin=p;
      while(*p && *p!=13)
        p++;
      k=p-pbegin;
      if (k) {
        if (j+k+2>=bytesLeft) {
          fprintf(stderr,
                  "CTAPI-CHERRY: Too many bytes received.\n");
          return KBD_HANDLER_ERROR_BUFFER_TOO_SMALL;
        }
        l=lineTags[i];
        *(d++)=l;
        *(d++)=k;
        j+=2;
        memmove(d, pbegin, k);
        j+=k;
        d+=k;
      }
      if (*p)
        p++;
    }
  } /* for */

  /* finalize data */
  if (j+3>=bytesLeft) {
    fprintf(stderr,
            "CTAPI-CHERRY: No room left for checksum.\n");
    return KBD_HANDLER_ERROR_ABORTED;
  }
  else {
    unsigned char m=0;

    /* add XOR checksum tag */
    *(d++)=0x8e;
    *(d++)=1;
    *d=0;
    j+=3;

    /* store size */
    dstart[1]=(unsigned char)((j-2) & 0xff);
    p=dstart;

    /* create XOR checksum */
    for (i=0; i<j; i++) {
      m=m^(*(p++));
    }
    /* store XOR checksum */
    *d=m;

    /* return real size of data */
    *buflen=j+sizeof(header);
  }

  return 0;
}



int kbd_handler_input_wait(KBD_HANDLER *kh, int ms, int outfd) {
  fd_set rset;
  struct timeval tv;
  int rv=0;

  while(1) {
    FD_ZERO(&rset);
    FD_SET(outfd, &rset);
    tv.tv_sec=0;
    tv.tv_usec=ms*1000; /* ms */
    rv=select(outfd+1, &rset, 0, 0, &tv);
    //fprintf(stderr, "Result of select: %d\n", rv);
    if (!(rv<0 && errno==EINTR))
      break;
  }
  if (rv<0) {
    fprintf(stderr, "ERROR: select(): %s\n", strerror(errno));
    return KBD_HANDLER_ERROR_ABORTED;
  }
  else if (rv==0) {
    //fprintf(stderr, "ERROR: Timeout.\n");
    return KBD_HANDLER_NO_CARD;
  }

  return 0;
}






